/*
    Histogram Equalization Filter for VirtualDub -- automatic contrast adjustment
    by means of histogram equalization. Copyright (C) 2004 Bernhard Lang

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  The author can be contacted at:
  Bernhard Lang
  B.Lang@fh-osnabrueck.de
*/

/* The framework of this filter is taken from
   Donald Graft's  "Histogram Equalization Filter for VirtualDub"
   (see http://neuron2.net/)
*/

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <crtdbg.h>
#include <math.h>
#include <stdlib.h>

#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#include "resource.h"
#include "filter.h"

///////////////////////////////////////////////////////////////////////////

int  RunProc(const FilterActivation *fa, const FilterFunctions *ff);
int  StartProc(FilterActivation *fa, const FilterFunctions *ff);
int  EndProc(FilterActivation *fa, const FilterFunctions *ff);
long ParamProc(FilterActivation *fa, const FilterFunctions *ff);
int  InitProc(FilterActivation *fa, const FilterFunctions *ff);
int  ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str);
void ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);

///////////////////////////////////////////////////////////////////////////

typedef struct MyFilterData {
  IFilterPreview  *ifp;
  int             strength;
  unsigned long   inHistogram[256];   // Input histogram
  unsigned long   outHistogram[256];  // Output histogram
  RECT            rInHisto;
  RECT            rOutHisto;
  long            inHistoMax;
  long            outHistoMax;
  int             disable;
} MyFilterData;

bool FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
  MyFilterData *mfd = (MyFilterData *)fa->filter_data;
  _snprintf(buf, buflen, "Config(%d)", mfd->strength);
  return true;
}

void ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
  FilterActivation *fa = (FilterActivation *)lpVoid;
  MyFilterData *mfd = (MyFilterData *)fa->filter_data;
  mfd->strength    = argv[0].asInt();                  // Read parameter as int
}

ScriptFunctionDef func_defs[]={
  { (ScriptFunctionPtr)ScriptConfig, "Config", "0i" }, // Specify one "int" parameter
  { NULL },
};

CScriptObject script_obj={
  NULL, func_defs
};

struct FilterDefinition filterDef_tutorial = {

  NULL, NULL, NULL,                                // next, prev, module
  "Easy_Histogram_Equalization_1_0",             // name
  "Perform global color histogram equalization.",  // desc
  "Bernhard Lang",                                 // maker
  NULL,                                            // private_data
  sizeof(MyFilterData),                            // inst_data_size

  InitProc,                                        // initProc
  NULL,                                            // deinitProc
  RunProc,                                         // runProc
  ParamProc,                                       // paramProc
  ConfigProc,                                      // configProc
  StringProc,                                      // stringProc
  NULL,                                            // startProc
  NULL,                                            // endProc

  &script_obj,                                     // script_obj
  FssProc,                                         // fssProc

};

long ParamProc(FilterActivation *fa, const FilterFunctions *ff) {

  fa->dst.offset = fa->src.offset;
    return 0;
}

int RunProc(const FilterActivation *fa, const FilterFunctions *ff) {

  MyFilterData *mfd = (MyFilterData *)fa->filter_data;
  PixDim w = fa->src.w;
  PixDim h = fa->src.h;
  Pixel32 *src, *dst;
  int x,y,i;

  // Initialize histogram tables
  for (i=0;i<256;i++) {
    mfd->inHistogram[i]=0;
    mfd->outHistogram[i]=0;
  }

  // Compute input histogram
  for (y=0;y<h;y++) {
    src = (Pixel32 *)((char *)fa->src.data + (sizeof(Pixel32)*w+fa->src.modulo)*y);
    for (x=0; x<w; x++) {
      mfd->inHistogram[(299*((*src>>16)&0xff)+587*((*src>>8)&0xff)+114*(*src&0xff))/1000]++;
      src++;
    }
  }

  // Compute transfer function from histogram
  double Exponent = double(mfd->strength)/255.0;
  double t[256];   // Transfer function LUT
  double tmp=0.0;  
  double weighted;
  for (i=0; i<256; i++) {
    if (Exponent>=0) weighted = pow(mfd->inHistogram[i],Exponent);
    else             weighted = 1.0/pow(mfd->inHistogram[i],-Exponent);
    t[i] = tmp+weighted/2;
    tmp += weighted;
  }
  // Normalize transfer function
  double offset=t[0];
  tmp=t[255]-offset;
  for (i=0; i<256; i++) {
    t[i] = (t[i]-offset)*255/tmp;
  }
  // Equalize image
  int Intensity;
  unsigned char red, green, blue;
  for(y=0; y<h; y++) {
    src = (Pixel32 *)((char *)fa->src.data + (sizeof(Pixel32)*w+fa->src.modulo)*y);
    dst = (Pixel32 *)((char *)fa->dst.data + (sizeof(Pixel32)*w+fa->dst.modulo)*y);
    for(x=0; x<w; x++) {
      // Get original color values
      red   = static_cast<unsigned char>((*src>>16)&0xff);
      green = static_cast<unsigned char>((*src>>8)&0xff);
      blue  = static_cast<unsigned char>(*src&0xff);
      src++;
      // Get maximum color value
      Intensity = (red>green) ? ((red>blue)?red:blue) : ((green>blue)?green:blue);
      // Equalize all three colors with same scale to avoid color shifting
      double scale = double(t[Intensity])/double(Intensity);
      red   = ( (unsigned int) (red   * scale ) );
      green = ( (unsigned int) (green * scale ) );
      blue  = ( (unsigned int) (blue  * scale ) );
      *dst++ = (red<<16) + (green<<8) + blue;
      // Compute output histogram
      mfd->outHistogram[(299*red+587*green+114*blue)/1000]++;
    }
  }
  return 0;
}

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_tutorial;

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
  if (!(fd_tutorial = ff->addFilter(fm, &filterDef_tutorial, sizeof(FilterDefinition))))
    return 1;

  vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
  vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

  return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
  ff->removeFilter(fd_tutorial);
}

int InitProc(FilterActivation *fa, const FilterFunctions *ff) {
  MyFilterData *mfd = (MyFilterData *)fa->filter_data;

  mfd->strength = 64;
  mfd->disable = 1;

  return 0;
}

static void levelsButtonCallback(bool fNewState, void *pvData)
{
  HWND hdlg = (HWND)pvData;

  EnableWindow(GetDlgItem(hdlg, IDC_SAMPLE), fNewState);
  if (fNewState == FALSE)
  {
    ShowWindow(GetDlgItem(hdlg, IDC_INHISTO), SW_SHOW);
    ShowWindow(GetDlgItem(hdlg, IDC_OUTHISTO), SW_SHOW);
  }
}

static void updateHistoDisplay(HWND hdlg, MyFilterData *mfd)
{
  int i;

  mfd->inHistoMax = mfd->inHistogram[0];
  for(i=1; i<256; i++)
    if ((long) mfd->inHistogram[i] > mfd->inHistoMax)
      mfd->inHistoMax = mfd->inHistogram[i];
  mfd->outHistoMax = mfd->outHistogram[0];
  for(i=1; i<256; i++)
    if ((long) mfd->outHistogram[i] > mfd->outHistoMax)
      mfd->outHistoMax = mfd->outHistogram[i];
  InvalidateRect(hdlg, &mfd->rInHisto, FALSE);
  InvalidateRect(hdlg, &mfd->rOutHisto, FALSE);
}

BOOL CALLBACK ConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
  MyFilterData *mfd = (MyFilterData *)GetWindowLong(hdlg, DWL_USER);

  switch(msg) {
    case WM_INITDIALOG:
      SetWindowLong(hdlg, DWL_USER, lParam);
      mfd = (MyFilterData *)lParam;

      GetWindowRect(GetDlgItem(hdlg, IDC_INHISTO), &mfd->rInHisto);
      ScreenToClient(hdlg, (POINT *)&mfd->rInHisto + 0);
      ScreenToClient(hdlg, (POINT *)&mfd->rInHisto + 1);
      GetWindowRect(GetDlgItem(hdlg, IDC_OUTHISTO), &mfd->rOutHisto);
      ScreenToClient(hdlg, (POINT *)&mfd->rOutHisto + 0);
      ScreenToClient(hdlg, (POINT *)&mfd->rOutHisto + 1);
      mfd->ifp->SetButtonCallback(levelsButtonCallback, (void *)hdlg);

      SendMessage(GetDlgItem(hdlg, IDC_STRENGTH), TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(0, 255));
      SendMessage(GetDlgItem(hdlg, IDC_STRENGTH), TBM_SETTICFREQ, 10 , 0);
      SendMessage(GetDlgItem(hdlg, IDC_STRENGTH), TBM_SETPOS, (WPARAM)TRUE, mfd->strength);
      SetDlgItemInt(hdlg, IDC_STRENGTHVAL, mfd->strength, FALSE);
      mfd->ifp->InitButton(GetDlgItem(hdlg, IDPREVIEW));

      return TRUE;

    case WM_COMMAND:
      switch(LOWORD(wParam)) {
      case IDPREVIEW:
        mfd->disable = 1;
        mfd->ifp->Toggle(hdlg);
        break;
      case IDOK:
        EndDialog(hdlg, 0);
        return TRUE;
       case IDHELP:
        {
        char prog[256];
        char path[256];
        LPTSTR ptr;

        GetModuleFileName(NULL, prog, 255);
        GetFullPathName(prog, 255, path, &ptr);
        *ptr = 0;
        strcat(path, "plugins\\Easy_Histeq.html");
        ShellExecute(hdlg, "open", path, NULL, NULL, SW_SHOWNORMAL);
        return TRUE;
        }
      case IDCANCEL:
        EndDialog(hdlg, 1);
        return TRUE;
      case IDC_SAMPLE:
        mfd->disable = 0;
        ShowWindow(GetDlgItem(hdlg, IDC_INHISTO), SW_HIDE);
        ShowWindow(GetDlgItem(hdlg, IDC_OUTHISTO), SW_HIDE);
        updateHistoDisplay(hdlg, mfd);
        break;
    }
    case WM_HSCROLL:
      {
        int strength;
        strength = SendMessage(GetDlgItem(hdlg, IDC_STRENGTH), TBM_GETPOS, 0, 0);
        if (strength != mfd->strength)
        {
          mfd->strength = strength;
          SetDlgItemInt(hdlg, IDC_STRENGTHVAL, strength, FALSE);
          mfd->ifp->RedoFrame();
          updateHistoDisplay(hdlg, mfd);
        }
        break;
      }
    case WM_PAINT:
    {
      HDC hdc;
      PAINTSTRUCT ps;
      RECT rPaint, rClip;
      int i;
      int x, xlo, xhi, w;
      long lMax;
      int xp, yp, h;
      long y;

      if (mfd->disable == 1)
      {
        return FALSE;
      }
      hdc = BeginPaint(hdlg, &ps);
      i = GetClipBox(hdc, &rClip);
      /* Input histogram. */
      if (i==ERROR || i==NULLREGION || IntersectRect(&rPaint, &mfd->rInHisto, &rClip))
      {
        lMax = mfd->inHistoMax;
        w = mfd->rInHisto.right - mfd->rInHisto.left;
        if (i==NULLREGION) {
          xlo = 0;
          xhi = w;
        } else {
          xlo = rPaint.left - mfd->rInHisto.left;
          xhi = rPaint.right - mfd->rInHisto.left;
        }
        FillRect(hdc, &mfd->rInHisto, (HBRUSH)GetStockObject(WHITE_BRUSH));
        for(x=xlo; x<xhi; x++)
        {
          i = (x * 0xFF00) / (w-1);

          if (i >= 0xFF00)
            y = mfd->inHistogram[255];

          else
            y = mfd->inHistogram[i>>8];

          xp = x+mfd->rInHisto.left;
          yp = mfd->rInHisto.bottom-1;
          h = MulDiv(y, mfd->rInHisto.bottom-mfd->rInHisto.top, lMax);

          if (h>0) {
            MoveToEx(hdc, x+mfd->rInHisto.left, yp, NULL);
            LineTo(hdc, x+mfd->rInHisto.left, yp - h);
          }
        }
      }
      /* Output histogram. */
      if (i==ERROR || i==NULLREGION || IntersectRect(&rPaint, &mfd->rOutHisto, &rClip))
      {
        lMax = mfd->outHistoMax;
        w = mfd->rOutHisto.right - mfd->rOutHisto.left;
        if (i==NULLREGION) {
          xlo = 0;
          xhi = w;
        } else {
          xlo = rPaint.left - mfd->rOutHisto.left;
          xhi = rPaint.right - mfd->rOutHisto.left;
        }
        FillRect(hdc, &mfd->rOutHisto, (HBRUSH)GetStockObject(WHITE_BRUSH));
        for(x=xlo; x<xhi; x++)
        {
          i = (x * 0xFF00) / (w-1);

          if (i >= 0xFF00)
            y = mfd->outHistogram[255];

          else
            y = mfd->outHistogram[i>>8];

          xp = x+mfd->rOutHisto.left;
          yp = mfd->rOutHisto.bottom-1;
          h = MulDiv(y, mfd->rOutHisto.bottom-mfd->rOutHisto.top, lMax);
          if (h>0) {
            MoveToEx(hdc, x+mfd->rOutHisto.left, yp, NULL);
            LineTo(hdc, x+mfd->rOutHisto.left, yp - h);
          }
        }
      }
      EndPaint(hdlg, &ps);
      break;
    }
  }
  return FALSE;
}

int ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd)
{
  MyFilterData *mfd = (MyFilterData *)fa->filter_data;
  MyFilterData mfd_old = *mfd;
  int ret;

  mfd->disable = 1;
  mfd->ifp = fa->ifp;

  if (DialogBoxParam(fa->filter->module->hInstModule, MAKEINTRESOURCE(IDD_FILTER),
    hwnd, ConfigDlgProc, (LPARAM)mfd))
  {
    *mfd = mfd_old;
    ret = TRUE;
  }
    else
  {
    ret = FALSE;
  }
  return(ret);
}

void StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str)
{
  MyFilterData *mfd = (MyFilterData *)fa->filter_data;

  sprintf(str, " (str %d)", mfd->strength);
}
